from datetime import date, datetime

from django.contrib.gis.geos import Point
from django.utils.encoding import force_text
from django.utils.functional import Promise


def round_geom(geom, precision):
    if isinstance(geom[0], (list, tuple)):
        return [round_geom(p, precision) for p in geom]

    return [round(l, precision) for l in geom]


def geom_accuracy(geom, point_accuracy=10):
    """
    Approximation of the 'radius' of a geometry in meters

    Done by taking the max distance between centroid and each of the coordinates.

    In case of a point, `point_accuracy` is used, which defaults to 10.
    """
    if geom.dims == 0:
        return point_accuracy
    elif geom.dims == 1:
        coords = geom.coords
    else:
        coords = geom.coords[0]
    centroid = geom.centroid
    return int(max([centroid.distance(Point(p)) for p in coords]) * 115200)


# simplified version of the implementation in
# https://github.com/jieter/django-tables2/blob/master/django_tables2/utils.py#L274
class Accessor(str):
    SEPARATOR = "."

    @property
    def bits(self):
        if self == "":
            return ()

        if self.SEPARATOR not in self:
            return self.split("__")

        return self.split(self.SEPARATOR)

    def resolve(self, context, safe=True):
        if isinstance(context, dict):
            return context[self]

        current = context
        for bit in self.bits:
            try:  # dictionary lookup
                current = current[bit]
            except (TypeError, AttributeError, KeyError):
                try:  # attribute lookup
                    current = getattr(current, bit)
                except (TypeError, AttributeError):
                    try:  # list-index lookup
                        current = current[int(bit)]
                    except (
                        IndexError,  # list index out of range
                        ValueError,  # invalid literal for int()
                        KeyError,  # dict without `int(bit)` key
                        TypeError,  # unsubscriptable object
                    ):
                        raise ValueError(
                            "Failed lookup for key [%s] in %r, " "when resolving the accessor %s" % (bit, current, self)
                        )

            if callable(current):
                if safe and getattr(current, "alters_data", False):
                    raise ValueError("refusing to call %s() because `.alters_data = True`" % repr(current))
                if not getattr(current, "do_not_call_in_templates", False):
                    current = current()
            # important that we break in None case, or a relationship
            # spanning across a null-key will raise an exception in the
            # next iteration, instead of defaulting.
            if current is None:
                break
        return current


def attr_or_key(obj, lookup):
    """Get an attribute if it exists on the object, else treat obj as dict"""
    return Accessor(lookup).resolve(obj)


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()

    if isinstance(obj, Promise):
        return force_text(obj)

    raise TypeError("Type %s not serializable" % type(obj))



from django.contrib.gis.gdal import SpatialReference
from django.contrib.gis.geos import Point
from django.utils.html import format_html


class PointFormatter(object):
    """Returns a nice HTML representation for a django.contrib.gis.geos.Point."""

    html_template = (
        '<span class="teramap-coordinates" title="{long_name}">'
        '<span class="teramap-coordinates-name text-muted">{short_name}</span>&nbsp;'
        '<span class="teramap-coordinates-coords">{point}</span>'
        "</span>"
    )

    INTEGER_FMT = "{p.x:0.0f} {p.y:0.0f}"
    FLOAT_FMT = "{p.x:0.4f}, {p.y:0.4f}"
    LATLNG_FLOAT_FMT = "{p.y:0.4f}, {p.x:0.4f}"

    CONFIG = {
        4326: {"short_name": "GPS", "template": LATLNG_FLOAT_FMT},
        28992: {"short_name": "RDS", "template": INTEGER_FMT},
        31370: {"short_name": "Lambert 1972", "template": INTEGER_FMT},
        "default": {"template": FLOAT_FMT},
    }

    def __init__(self, p, srs=None):
        """
        arguments:
            p (Point): point to display.
            srs (SpatialReference): If not None, point will be tranformed according to srs and displayed using srs.
        """
        if not isinstance(p, Point):
            p = Point(p)

        if not isinstance(srs, SpatialReference):
            try:
                srs = SpatialReference(srs)
            except Exception:
                srs = None

        if srs is not None:
            p = p.transform(srs, clone=True)

        self.p = p

    @property
    def config(self):
        return self.CONFIG.get(self.p.srid, self.CONFIG["default"])

    @property
    def point_template(self):
        return self.config.get("template", self.CONFIG["default"]["template"])

    @property
    def short_name(self):
        return self.config.get("short_name", self.p.srs.name)

    def __str__(self):
        return format_html(
            self.html_template,
            long_name=self.p.srs.name,
            short_name=self.short_name,
            point=self.point_template.format(p=self.p),
        )


def format_point(p, srs=None):
    """Short way to use the PointFormatter class"""
    if p is None or (p.srs is None and srs is None):
        return ""
    return str(PointFormatter(p, srs))



from django import template
from django.contrib.staticfiles.templatetags.staticfiles import static

from .. import __version__
from ..formatters import format_point

register = template.Library()


STATIC_FILE_FMT = "{}?version={}"


@register.simple_tag
def teramap_version():
    return __version__


@register.simple_tag
def teramap_js():
    return STATIC_FILE_FMT.format(static("teramap/teramap.js"), __version__)


@register.simple_tag
def teramap_css():
    return STATIC_FILE_FMT.format(static("teramap/teramap.css"), __version__)


@register.filter
def driving_directions(point):
    """
    Output a link with directions to point

    Use links to apple maps because it will redirect to native apps on Linux/Android and Windows.
    """
    return "https://maps.apple.com/?daddr={point.y},{point.x}".format(point=point)


register.filter(format_point)



import itertools
import json

from django.conf import settings
from django.urls import reverse
from django.utils.safestring import mark_safe

from .utils import Accessor, attr_or_key, json_serial, round_geom

# Assume 1 degree = 115 km = 115000 m, simplify to 25m
DEFAULT_SIMPLIFY = getattr(settings, "GEOJSON_DEFAULT_SIMPLIFY", 25 / 115000)


class FeatureBase(object):
    def __len__(self):
        return len(self.features)

    def len(self):
        # length function to use in templates
        return self.__len__()

    def union(self, other):
        key = "collection_properties"
        properties = {}
        properties.update(getattr(self, key, None) or {})
        properties.update(getattr(other, key, None) or {})

        return FeatureCollection(
            itertools.chain(self.features, other.features),
            collection_properties=properties if len(properties) > 0 else None,
        )

    # support adding two Features/FeatureCollections:
    # geojson = geojson_a + geojson_b
    # geojson += a
    __add__ = union
    __iadd__ = union


class FeatureCollection(FeatureBase):
    def __init__(self, features, collection_properties=None):
        self.features = list(features)
        self.collection_properties = collection_properties

    def to_geojson(self):
        ret = {
            "type": "FeatureCollection",
            "features": [
                feature.to_geojson() if isinstance(feature, Feature) else feature for feature in self.features
            ],
        }
        if getattr(self, "collection_properties", None):
            ret["properties"] = self.collection_properties

        return ret

    def to_geojson_str(self, indent=None):
        return json.dumps(self.to_geojson(), indent=indent, default=json_serial)

    def __str__(self):
        return mark_safe(str(self.to_geojson_str()))


class Feature(FeatureBase):
    @property
    def features(self):
        return (self,)

    def to_geojson(self):
        return {
            "type": "Feature",
            "properties": self.properties,
            "geometry": {"type": self.geom_type, "coordinates": self.coords},
        }


class Point(Feature):
    geom_type = "Point"

    def __init__(self, coords, properties=None):
        self.coords = coords
        self.properties = properties or {}


class LineString(Point):
    geom_type = "LineString"


class TransformToPoint(object):
    """
    Mixin to convert non-point geometries to a point.

    Useful to present lots of polygons as markers to reduce page load time.
    """

    def geometry(self, obj):
        geom = super(TransformToPoint, self).geometry(obj)

        if geom.geom_typeid > 0:
            geom = geom.centroid
        return geom


class GeoJSON(FeatureCollection):
    """
    Transform a list of Models to GeoJSON with properties according to specification.

    Arguments:
        data: list of dicts or QuerySet.
        extra_properties: dict of the same structure as the properties property to add to get_properties
            dict for this instance.
        geom_field: name of the field to use as the geometry, overriding the `geom_field` attribute.
    """

    properties = {"id": "pk"}
    geom_field = "geom"
    precision = 6
    simplify = DEFAULT_SIMPLIFY

    def __init__(self, data, extra_properties=None, geom_field=None):
        self.data = data or []

        self.extra_properties = extra_properties or {}

        if geom_field is not None:
            self.geom_field = geom_field

    def get_properties(self, obj):
        properties = self.properties
        properties.update(self.extra_properties)

        ret = {}

        for key, lookup in properties.items():
            if isinstance(lookup, bool):
                ret[key] = lookup
            elif callable(lookup):
                ret[key] = lookup(obj)
            else:
                ret[key] = Accessor(lookup).resolve(obj)

        return ret

    def geometry(self, obj):
        """
        Return the geometry simplified according to the current simplification setting.
        """
        geom = Accessor(self.geom_field).resolve(obj)
        return geom.simplify(self.simplify, preserve_topology=True)

    def feature(self, obj):
        try:
            geom = self.geometry(obj)
        except AttributeError:
            return None

        return {
            "type": "Feature",
            "properties": self.get_properties(obj),
            "geometry": {"type": geom.geom_type, "coordinates": round_geom(geom.coords, self.precision)},
        }

    @property
    def features(self):
        """
        Return the GeoJSON representation of the list of locations.
        """
        if not hasattr(self, "_features"):
            self._features = []
            for location in self.data:
                feature = self.feature(location)
                if feature:
                    self._features.append(feature)

        return self._features

    def url(self, obj):
        return reverse(self.url_name, kwargs={"pk": attr_or_key(obj, "pk")})



from django.http import JsonResponse
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.gzip import gzip_page

from .geojson import GeoJSON


class GeoJSONView(View):
    """
    Return a GeoJSON for the object list.

    Example:

    class View(teramap.GeoJSONView, ListView):
        GeoJSONClass = LocationGeoJSON
        model = Location

    """

    GeoJSONClass = GeoJSON

    def __init__(self, *args, **kwargs):
        class Klass(self.GeoJSONClass):
            pass

        if hasattr(self, "geom_field"):
            Klass.geom_field = self.geom_field
        if hasattr(self, "get_properties"):
            Klass.get_properties = lambda obj: self.get_properties(obj)

        self.geojson = Klass([])

        super(GeoJSONView, self).__init__(*args, **kwargs)

    @method_decorator(gzip_page)
    def dispatch(self, *args, **kwargs):
        return super(GeoJSONView, self).dispatch(*args, **kwargs)

    def get_data(self, context):
        try:
            return context["object_list"]
        except KeyError:
            pass

        try:
            return [context["object"]]
        except KeyError:
            pass

        return []

    def render_to_response(self, context, **response_kwargs):
        self.geojson.data = self.get_data(context)

        return JsonResponse(self.geojson.to_geojson())
